%{
#include "bison2.h"
#include "re2c.h"
#include "vsm.h"
#include "basic.h"


//---------------------------------------------------------------------------
// #define YYDEBUG				1
// #define YYERROR_VERBOSE		1
// int yydebug = 1;


%}

%token TOKEN_IF
%token TOKEN_THEN
%token TOKEN_GOTO
%token TOKEN_ELSE
%token TOKEN_FOR
%token TOKEN_TO
%token TOKEN_NEXT
%token TOKEN_PRINT
%token TOKEN_GOSUB
%token TOKEN_RETURN
%token TOKEN_CLS
%token TOKEN_PSET
%token TOKEN_LINE
%token TOKEN_CIRCLE
%token TOKEN_DIM
%token TOKEN_RND
%token TOKEN_REM
%token TOKEN_END
%token TOKEN_PLUS
%token TOKEN_MINUS
%token TOKEN_MULT
%token TOKEN_DIV
%token TOKEN_MOD
%token TOKEN_GE
%token TOKEN_GT
%token TOKEN_LE
%token TOKEN_LT
%token TOKEN_EQ
%token TOKEN_NE
%token TOKEN_LPAR
%token TOKEN_RPAR
%token TOKEN_COMMA
%token TOKEN_SEMI
%token TOKEN_COLON
%token TOKEN_NOT
%token TOKEN_AND
%token TOKEN_OR
%token TOKEN_XOR
%token TOKEN_NUM
%token TOKEN_VAR
%token TOKEN_STR
%token TOKEN_LF
%token TOKEN_ERR

%expect 3

%%
program		: program line
            | line
            ;

line        : TOKEN_NUM
				{
					u16 adr = _Atoi($1);

					if(Bison.adrCnt > adr)
					{
						yyerror("adrCnt Err!");
						YYERROR;
					}
					else if(VsmCodeIsFull() == TRUE)
					{
						yyerror("CodeBuf Full!");
						YYERROR;
					}
					else if(VsmWorkIsFull() == TRUE)
					{
						yyerror("WorkBuf Full!");
						YYERROR;
					}

					Bison.adrCnt = adr;
					VsmCodeAddAdr(adr);
				}
              body TOKEN_LF
				{
					Re2cBufClear();
				}
			| TOKEN_LF
            ;

body        : body TOKEN_COLON stmt
            | stmt
            ;

stmt        : stmt_for
            | stmt_next
            | stmt_print
            | stmt_gosub
            | stmt_return
            | stmt_rem
            | stmt_end
            | stmt_goto
            | stmt_if
            | stmt_pset
            | stmt_line
            | stmt_circle
            | stmt_cls
            | stmt_dim
            | stmt_var
            ;

stmt_end    : TOKEN_END
				{
					VsmCodeAddEnd();
				}
            ;

stmt_next   : TOKEN_NEXT
				{
					VsmCodeAddNext();

					if(BisonIsJmpBufEmpty() == TRUE)
					{
						yyerror("JmpBuf Empty!");
						YYERROR;
					}
					VsmCodePatchFor(BisonPopJmpBuf());
				}
            ;

stmt_return : TOKEN_RETURN
				{
					VsmCodeAddReturn();
				}
            ;

stmt_rem    : TOKEN_REM
				{
					Re2cSkipLF();
				}
            ;

stmt_cls    : TOKEN_CLS expr
				{
					VsmCodeAddCls();
				}
            ;

stmt_gosub  : TOKEN_GOSUB expr
				{
					VsmCodeAddGosub();
				}
            ;

stmt_goto   : TOKEN_GOTO expr
				{
					VsmCodeAddGoto();
				}
            ;

stmt_if     : stmt_if2
				{
					VsmCodePatchBeq(BisonPopJmpBuf());
				}
            | stmt_if2 TOKEN_ELSE
				{
					VsmCodeAddJmp();
					VsmCodePatchBeq(BisonPopJmpBuf());

					if(BisonIsJmpBufFull() == TRUE)
					{
						yyerror("JmpBuf Full!");
						YYERROR;
					}
					BisonPushJmpBuf();
				}
              body
				{
					VsmCodePatchJmp(BisonPopJmpBuf());
				}
            ;

stmt_if2    : TOKEN_IF cond TOKEN_THEN
				{
					VsmCodeAddBeq();

					if(BisonIsJmpBufFull() == TRUE)
					{
						yyerror("Bison.JmpBuf Full!");
						YYERROR;
					}
					BisonPushJmpBuf();
				}
              body
            ;

stmt_pset   : TOKEN_PSET TOKEN_LPAR expr TOKEN_COMMA expr TOKEN_RPAR TOKEN_COMMA expr
				{
					VsmCodeAddPset();
				}
            ;

stmt_line   : TOKEN_LINE TOKEN_LPAR expr TOKEN_COMMA expr TOKEN_RPAR TOKEN_MINUS TOKEN_LPAR expr TOKEN_COMMA expr TOKEN_RPAR TOKEN_COMMA expr
				{
					VsmCodeAddLine();
				}
            ;

stmt_circle : TOKEN_CIRCLE TOKEN_LPAR expr TOKEN_COMMA expr TOKEN_RPAR TOKEN_COMMA expr TOKEN_COMMA expr
				{
					VsmCodeAddCircle();
				}
            ;

stmt_for    : TOKEN_FOR TOKEN_VAR TOKEN_EQ expr
				{
					VsmCodeAddPushVar($2);
					VsmCodeAddForInit();
				}
			  TOKEN_TO expr
				{
					VsmCodeAddFor();

					if(BisonIsJmpBufFull() == TRUE)
					{
						yyerror("Bison.JmpBuf Full!");
						YYERROR;
					}
					BisonPushJmpBuf();
				}
            ;

stmt_print  : TOKEN_PRINT stmt_print2
				{
					VsmCodeAddPrintLf();
				}
            ;

stmt_print2 : stmt_print2 TOKEN_COMMA stmt_print3
            | stmt_print3
            ;

stmt_print3 : TOKEN_STR
				{
					VsmCodeAddPushStr($1);
					VsmCodeAddPrint();
				}
            | expr
				{
					VsmCodeAddPrint();
				}
            ;

stmt_dim    : TOKEN_DIM stmt_dim2
            ;

stmt_dim2   : stmt_dim2 TOKEN_COMMA stmt_dim3
            | stmt_dim3
            ;

stmt_dim3   : TOKEN_VAR TOKEN_LPAR expr TOKEN_RPAR
				{
					VsmCodeAddPushArray($1);
					VsmCodeAddDim();
				}
            ;

stmt_var    : TOKEN_VAR TOKEN_EQ expr
				{
					VsmCodeAddPushVar($1);
					VsmCodeAddVar();
				}
            | TOKEN_VAR TOKEN_LPAR expr TOKEN_RPAR TOKEN_EQ expr
				{
					VsmCodeAddPushArray($1);
					VsmCodeAddArray();
				}
            ;

cond        : expr TOKEN_GT expr
				{
					VsmCodeAddGt();
				}
            | expr TOKEN_GE expr
				{
					VsmCodeAddGe();
				}
            | expr TOKEN_LT expr
				{
					VsmCodeAddLt();
				}
            | expr TOKEN_LE expr
				{
					VsmCodeAddLe();
				}
            | expr TOKEN_NE expr
				{
					VsmCodeAddNe();
				}
            | expr TOKEN_EQ expr
				{
					VsmCodeAddEq();
				}
            ;

expr        : expr TOKEN_PLUS  term
				{
					VsmCodeAddPlus();
				}
            | expr TOKEN_MINUS term
				{
					VsmCodeAddMinus();
				}
            | term
            ;

term        : term TOKEN_MULT factor
				{
					VsmCodeAddMult();
				}
            | term TOKEN_DIV  factor
				{
					VsmCodeAddDiv();
				}
            | term TOKEN_MOD  factor
				{
					VsmCodeAddMod();
				}
            | factor
            ;

factor      : TOKEN_NUM
				{
					VsmCodeAddPushNum(_Atoi($1));
				}
            | TOKEN_VAR
				{
					VsmCodeAddPushVar($1);
				}
            | TOKEN_VAR TOKEN_LPAR expr TOKEN_RPAR
				{
					VsmCodeAddPushArray($1);
				}
            | TOKEN_LPAR expr TOKEN_RPAR
            | rnd
            ;

rnd         : TOKEN_RND TOKEN_LPAR expr TOKEN_RPAR
				{
					VsmCodeAddRnd();
				}
            ;

%%

//---------------------------------------------------------------------------
void BisonInit()
{
	_Memset(&Bison, 0x00, sizeof(ST_BISON));
}
//---------------------------------------------------------------------------
bool BisonIsJmpBufFull()
{
	return (Bison.jmpCnt >= BISON_MAX_JMP_CNT) ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
bool BisonIsJmpBufEmpty()
{
	return (Bison.jmpCnt == 0) ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
void BisonPushJmpBuf()
{
	Bison.jmpBuf[ Bison.jmpCnt++ ] = VsmCodeGetCnt() - 1;
}
//---------------------------------------------------------------------------
u16 BisonPopJmpBuf()
{
	_ASSERT(Bison.jmpCnt > 0);
	return Bison.jmpBuf[ --Bison.jmpCnt ];
}
//---------------------------------------------------------------------------
void yyerror(char* pStr)
{
	BasicDrawErrStr(pStr, Bison.adrCnt);
}
