#ifndef __VSM_H__
#define __VSM_H__


#include "libgba/gba.h"


//---------------------------------------------------------------------------
#define VSM_MAX_STACK_BUF_CNT			100
#define VSM_MAX_CODE_BUF_CNT			1000
#define VSM_MAX_WORK_BUF_CNT			2000
#define VSM_MAX_VAR_BUF_CNT				200
#define VSM_MAX_ADR_BUF_CNT				200

#define VSM_MAX_RUN_CNT					50

enum {
	CODE_END = 0x01,
	CODE_NEXT,
	CODE_RETURN,
	CODE_CLS,
	CODE_GOSUB,			// 5
	CODE_GOTO,
	CODE_RND,
	CODE_BEQ,
	CODE_JMP,
	CODE_PSET,			// 10
	CODE_LINE,
	CODE_CIRCLE,
	CODE_FOR_INIT,
	CODE_FOR,
	CODE_PRINT,			// 15
	CODE_PRINT_LF,
	CODE_VAR,
	CODE_ARRAY,
	CODE_DIM,
	CODE_ADR,			// 20
	CODE_GT,
	CODE_GE,
	CODE_LT,
	CODE_LE,
	CODE_NE,			// 25
	CODE_EQ,
	CODE_PLUS,
	CODE_MINUS,
	CODE_MULT,
	CODE_DIV,			// 30
	CODE_MOD,
	CODE_PUSH_NUM,
	CODE_PUSH_VAR,
	CODE_PUSH_STR,
	CODE_PUSH_ARRAY,	// 35
};


enum {
	MODE_EXEC = 0x01,
	MODE_ERR,
	MODE_END,
};

enum {
	STACK_NONE = 0x01,
	STACK_NUM,
	STACK_VAR,
	STACK_ARRAY,
	STACK_STR,
	STACK_FOR,
	STACK_GOSUB,
};

enum {
	VAR_NONE = 0x01,
	VAR_VAR,
	VAR_ARRAY,
};

//---------------------------------------------------------------------------
typedef struct {
	u16 type;			// VAR_VAR or VAR_ARRAY
	s16 size;			// ϐ̏ꍇ1Az̏ꍇ2ȏ
	u16 strAdr;			// workBuf̖Ȍꏊ
	u16 varAdr;			// workBuf̒l̏ꏊ
} ST_VSM_VAR_BUF;


typedef struct {
	u16 adr;			// Ԓn
	u16 pc;				// codeBuf̏ꏊ
} ST_VSM_ADR_BUF;


typedef struct {
	// vm
	u16  mode;
	s16  pc;			// ĂAhX
	u16  cnt;			// s̃AhXJE^
	bool isJmp;

	// stack
	s16 stackCnt;
	u16 stackBuf[VSM_MAX_STACK_BUF_CNT];

	// code
	s16 codeCnt;
	u16 codeBuf[VSM_MAX_CODE_BUF_CNT];

	// workiϐAzAi[܂j
	s16 workCnt;
	u16 workBuf[VSM_MAX_WORK_BUF_CNT];

	// variϐAz̏i[܂j
	s16 varCnt;
	ST_VSM_VAR_BUF varBuf[VSM_MAX_VAR_BUF_CNT];

	// adriGOSUBAGOTO߂Ŏg܂j
	s16 adrCnt;
	ST_VSM_ADR_BUF adrBuf[VSM_MAX_ADR_BUF_CNT];

} ST_VSM;


//---------------------------------------------------------------------------
void VsmInit();
void VsmReset();
void VsmExec();
bool VsmIsExecEnd();


//----
void VsmCodeAdd(u16 code);

void VsmCodeAddEnd();
void VsmCodeAddNext();
void VsmCodeAddReturn();
void VsmCodeAddCls();
void VsmCodeAddGosub();
void VsmCodeAddGoto();
void VsmCodeAddRnd();
void VsmCodeAddBeq();
void VsmCodeAddJmp();
void VsmCodeAddPset();
void VsmCodeAddLine();
void VsmCodeAddCircle();
void VsmCodeAddForInit();
void VsmCodeAddFor();
void VsmCodeAddPrint();
void VsmCodeAddPrintLf();
void VsmCodeAddVar();
void VsmCodeAddArray();
void VsmCodeAddDim();
void VsmCodeAddAdr(u16 adr);

void VsmCodeAddGt();
void VsmCodeAddGe();
void VsmCodeAddLt();
void VsmCodeAddLe();
void VsmCodeAddNe();
void VsmCodeAddEq();

void VsmCodeAddPlus();
void VsmCodeAddMinus();
void VsmCodeAddMult();
void VsmCodeAddDiv();
void VsmCodeAddMod();

void VsmCodeAddPushNum(u16 num);
void VsmCodeAddPushVar(char* pStr);
void VsmCodeAddPushArray(char* pStr);
void VsmCodeAddPushStr(char* pStr);

void VsmCodePatchBeq(u16 tmpCnt);
void VsmCodePatchJmp(u16 tmpCnt);
void VsmCodePatchFor(u16 tmpCnt);

u16  VsmCodeGetCnt();
bool VsmCodeIsFull();


//----
u16  VsmWorkAddStr(char* pStr);
u16  VsmWorkAddVar(u16 size);

u16  VsmWrokGetAdr(char* pStr);
bool VsmWorkIsFull();
bool VsmWorkIsReg(char* pStr);


//----
void  VsmStackPush(u16 num, u16 type);
u16   VsmStackPop();
u16   VsmStackPopNum();
char* VsmStackPopStr();
u16   VsmStackPopAdr();

u16   VsmStackGetTopType();
u16   VsmStackGetTopNum();
bool  VsmStackIsFull();


//----
u16  VsmVarReg(u16 strAdr, u16 size);
u16  VsmVarAdd(u16 strAdr);

u16  VsmVarGetAdr(u16 strAdr, u16 type);
u16  VsmVarGetSize(u16 strAdr, u16 type);
bool VsmVarIsReg(u16 strAdr, u16 type);
bool VsmVarIsFull();


//----
void VsmAdrAdd(u16 adr, u16 pc);
u16  VsmAdrGetPc(u16 adr);
bool VsmAdrIsPc(u16 adr);
bool VsmAdrIsFull();


//----
void VsmDebugCodeBuf();
void VsmDebugWorkBuf();
void VsmDebugAdrBuf();


#endif
