#include "basic.h"
#include "vsm.h"


//---------------------------------------------------------------------------
ST_VSM Vsm;


//---------------------------------------------------------------------------
void VsmInit()
{
	TRACEOUT("Init\n");

	_Memset(&Vsm, 0x00, sizeof(ST_VSM));
}
//---------------------------------------------------------------------------
void VsmReset()
{
	VsmDebugCodeBuf();
	VsmDebugWorkBuf();
	VsmDebugAdrBuf();

	TRACEOUT("------------------------------\n"); 
	TRACEOUT("Exec\n");

	Vsm.mode = MODE_EXEC;
}
//---------------------------------------------------------------------------
void VsmExec()
{
	_ASSERT(Vsm.mode == MODE_EXEC);

	u16  op;
	u16  arg1;
	u16  arg2;
	u16  arg3;
	u16  arg4;
	u16  arg5;

	u16 i;
	for(i=0; i<VSM_MAX_RUN_CNT; i++)
	{
		if(VsmWorkIsFull() == TRUE)
		{
			BasicDrawErrStr("WorkBuf Full!", Vsm.cnt);
			goto Err;
		}
		else if(VsmStackIsFull() == TRUE)
		{
			BasicDrawErrStr("StackBuf Full!", Vsm.cnt);
			goto Err;
		}
		else if(VsmVarIsFull() == TRUE)
		{
			BasicDrawErrStr("VarBuf Full!", Vsm.cnt);
			goto Err;
		}

		if(Vsm.pc >= Vsm.codeCnt)
		{
			goto End;
		}

		op = Vsm.codeBuf[Vsm.pc++];

		switch(op)
		{
		case CODE_END:
			TRACEOUT("End\n");
			goto End;
			break;

		case CODE_NEXT:
			TRACEOUT("Next\n");
			arg1 = VsmStackPop();				// type
			arg2 = VsmStackPop();

			if(arg1 != STACK_FOR)
			{
				TRACEOUT("Next Err!\n");
				goto Err;
			}

			arg3 = VsmStackPopNum();			// var(workAdr)
			arg4 = VsmStackPopNum();			// pc
			Vsm.workBuf[arg3]++;
			Vsm.pc = arg4;

			VsmStackPush(STACK_NUM, arg4);		// pc
			VsmStackPush(STACK_NUM, arg3);		// var(workAdr)
			break;

		case CODE_RETURN:
			TRACEOUT("Return\n");

			arg1 = VsmStackPop();				// gosub
			arg2 = VsmStackPop();

			if(arg1 != STACK_GOSUB)
			{
				TRACEOUT("Return Err!\n");
				goto Err;
			}
			Vsm.pc = VsmStackPopNum();
			break;

		case CODE_CLS:
			TRACEOUT("Cls\n");
			arg1 = VsmStackPopNum();
			break;

		case CODE_GOSUB:
			TRACEOUT("Gosub\n");
			arg1 = VsmStackPopNum();

			if(VsmAdrIsPc(arg1) == FALSE)
			{
				TRACEOUT("Gosub Err!\n");
				goto Err;
			}

			VsmStackPush(STACK_NUM, Vsm.pc);	// pc
			VsmStackPush(STACK_GOSUB, 0);		// gosub
			Vsm.pc = VsmAdrGetPc(arg1);
			break;

		case CODE_GOTO:
			TRACEOUT("Goto\n");
			arg1 = VsmStackPopNum();

			if(VsmAdrIsPc(arg1) == FALSE)
			{
				TRACEOUT("Gosub Err!\n");
				goto Err;
			}
			Vsm.pc = VsmAdrGetPc(arg1);
			break;

		case CODE_RND:
			TRACEOUT("Rnd\n");
			arg1 = VsmStackPopNum();
			arg2 = BasicGetRand(arg1);

			VsmStackPush(STACK_NUM, arg2);
			break;

		case CODE_BEQ:
			TRACEOUT("Beq\n");

			if(Vsm.isJmp == TRUE)
			{
				Vsm.pc = Vsm.codeBuf[Vsm.pc];
			}
			else
			{
				Vsm.pc++;
			}
			break;

		case CODE_JMP:
			TRACEOUT("Jmp\n");
			Vsm.pc = Vsm.codeBuf[Vsm.pc];
			break;

		case CODE_PSET:
			TRACEOUT("Pset\n");
			arg1 = VsmStackPopNum();
			arg2 = VsmStackPopNum();
			arg3 = VsmStackPopNum();
			BasicDrawPset(arg3, arg2, arg1);
			break;

		case CODE_LINE:
			TRACEOUT("Line\n");
			arg1 = VsmStackPopNum();
			arg2 = VsmStackPopNum();
			arg3 = VsmStackPopNum();
			arg4 = VsmStackPopNum();
			arg5 = VsmStackPopNum();
			BasicDrawLine(arg5, arg4, arg3, arg2, arg1);
			break;

		case CODE_CIRCLE:
			TRACEOUT("Circle\n");
			arg1 = VsmStackPopNum();
			arg2 = VsmStackPopNum();
			arg3 = VsmStackPopNum();
			arg4 = VsmStackPopNum();
			BasicDrawCircle(arg4, arg3, arg2, arg1);
			break;

		case CODE_FOR_INIT:
			TRACEOUT("ForInit\n");
			arg1 = VsmStackPopAdr();
			arg2 = VsmStackPopNum();
			Vsm.workBuf[arg1] = arg2;

			VsmStackPush(STACK_NUM, Vsm.pc);	// pc
			VsmStackPush(STACK_NUM, arg1);		// var
			break;

		case CODE_FOR:
			TRACEOUT("For\n");
			arg1 = VsmStackPopNum();			// endCnt
			arg2 = VsmStackPopNum();			// var(workAdr)
			arg3 = VsmStackPopNum();			// pc

			TRACEOUT("%d > %d\n", Vsm.workBuf[arg2], arg1);

			if(Vsm.workBuf[arg2] > arg1)
			{
				Vsm.pc = Vsm.codeBuf[Vsm.pc];
			}
			else
			{
				VsmStackPush(STACK_NUM, arg3);	// pc
				VsmStackPush(STACK_NUM, arg2);	// var(workAdr)
				VsmStackPush(STACK_FOR, 0);		// for
				Vsm.pc++;
			}
			break;

		case CODE_PRINT:
			TRACEOUT("Print\n");
			arg1 = VsmStackGetTopType();

			switch(arg1)
			{
			case STACK_NUM:
			case STACK_VAR:
			case STACK_ARRAY:
				BasicDrawNum((s16)VsmStackPopNum());
				break;

			case STACK_STR:
				BasicDrawStr((char*)VsmStackPopStr());
				break;

			default:
				TRACEOUT("Illigal Stack!\n");
				goto Err;
				break;
			}
			break;

		case CODE_PRINT_LF:
			TRACEOUT("PrintLf\n");
			BasicDrawLf();
			break;

		case CODE_VAR:
			TRACEOUT("Var\n");
			arg1 = VsmStackPopAdr();
			arg2 = VsmStackPopNum();
			Vsm.workBuf[arg1] = arg2;
			break;

		case CODE_ARRAY:
			TRACEOUT("Array\n");
			arg1 = VsmStackPop();			// Array
			arg2 = VsmStackPop();
			arg3 = VsmStackPopNum();		// l
			arg4 = VsmStackPopNum();		// Arrayʒu

			if(VsmVarIsReg(arg2, VAR_ARRAY) == FALSE)
			{
				BasicDrawErrStr("Array Reg err!", Vsm.cnt);
				goto Err;
			}
			else if(VsmVarGetSize(arg2, VAR_ARRAY) <= arg4)
			{
				BasicDrawErrStr("Array Size err!", Vsm.cnt);
				goto Err;
			}
			Vsm.workBuf[VsmVarGetAdr(arg2, VAR_ARRAY) + arg4] = arg3;
			break;

		case CODE_DIM:
			TRACEOUT("Dim\n");
			arg1 = VsmStackPop();
			arg2 = VsmStackPop();
			arg3 = VsmStackPopNum();

			if(VsmVarIsReg(arg2, VAR_ARRAY) == TRUE)
			{
				BasicDrawErrStr("Dim err!\n", Vsm.cnt);
				goto Err;
			}
			VsmVarReg(arg2, arg3);
			break;

		case CODE_ADR:
			Vsm.cnt = Vsm.codeBuf[Vsm.pc++];

			TRACEOUT("------------------------------\n"); 
			TRACEOUT("Adr %d\n", Vsm.cnt);
			break;

		case CODE_GT:
			TRACEOUT("Gt\n");
			arg1 = VsmStackPopNum();
			arg2 = VsmStackPopNum();
			Vsm.isJmp = (arg2 > arg1) ? FALSE : TRUE;

			TRACEOUT("%d > %d\n", arg2, arg1);
			break;

		case CODE_GE:
			TRACEOUT("Ge\n");
			arg1 = VsmStackPopNum();
			arg2 = VsmStackPopNum();
			Vsm.isJmp = (arg2 >= arg1) ? FALSE : TRUE;

			TRACEOUT("%d >= %d\n", arg2, arg1);
			break;

		case CODE_LT:
			TRACEOUT("Lt\n");
			arg1 = VsmStackPopNum();
			arg2 = VsmStackPopNum();
			Vsm.isJmp = (arg2 < arg1) ? FALSE : TRUE;

			TRACEOUT("%d < %d\n", arg2, arg1);
			break;

		case CODE_LE:
			TRACEOUT("Le\n");
			arg1 = VsmStackPopNum();
			arg2 = VsmStackPopNum();
			Vsm.isJmp = (arg2 <= arg1) ? FALSE : TRUE;

			TRACEOUT("%d <= %d\n", arg2, arg1);
			break;

		case CODE_NE:
			TRACEOUT("Ne\n");
			arg1 = VsmStackPopNum();
			arg2 = VsmStackPopNum();
			Vsm.isJmp = (arg2 != arg1) ? FALSE : TRUE;

			TRACEOUT("%d <> %d\n", arg2, arg1);
			break;

		case CODE_EQ:
			TRACEOUT("Eq\n");
			arg1 = VsmStackPopNum();
			arg2 = VsmStackPopNum();
			Vsm.isJmp = (arg2 == arg1) ? FALSE : TRUE;

			TRACEOUT("%d == %d\n", arg2, arg1);
			break;

		case CODE_PLUS:
			TRACEOUT("Plus\n");
			arg1 = VsmStackPopNum();
			arg2 = VsmStackPopNum();
			VsmStackPush(STACK_NUM, (s16)arg2 + (s16)arg1);
			break;

		case CODE_MINUS:
			TRACEOUT("Minus\n");
			arg1 = VsmStackPopNum();
			arg2 = VsmStackPopNum();
			VsmStackPush(STACK_NUM, (s16)arg2 - (s16)arg1);
			break;

		case CODE_MULT:
			TRACEOUT("Mult\n");
			arg1 = VsmStackPopNum();
			arg2 = VsmStackPopNum();
			VsmStackPush(STACK_NUM, (s16)arg2 * (s16)arg1);
			break;

		case CODE_DIV:
			TRACEOUT("Div\n");
			arg1 = VsmStackPopNum();
			arg2 = VsmStackPopNum();
			VsmStackPush(STACK_NUM, Div((s16)arg2, (s16)arg1));
			break;

		case CODE_MOD:
			TRACEOUT("Mod\n");
			arg1 = VsmStackPopNum();
			arg2 = VsmStackPopNum();
			VsmStackPush(STACK_NUM, DivMod((s16)arg2, (s16)arg1));
			break;

		case CODE_PUSH_NUM:
			TRACEOUT("PushNum\n");
			VsmStackPush(STACK_NUM, Vsm.codeBuf[Vsm.pc]);
			Vsm.pc++;
			break;

		case CODE_PUSH_VAR:
			TRACEOUT("PushVar\n");
			VsmStackPush(STACK_VAR, Vsm.codeBuf[Vsm.pc]);
			Vsm.pc++;
			break;

		case CODE_PUSH_STR:
			TRACEOUT("PushStr\n");
			VsmStackPush(STACK_STR, Vsm.codeBuf[Vsm.pc]);
			Vsm.pc++;
			break;

		case CODE_PUSH_ARRAY:
			TRACEOUT("PushArray\n");
			VsmStackPush(STACK_ARRAY, Vsm.codeBuf[Vsm.pc]);
			Vsm.pc++;
			break;

		default:
			BasicDrawErrStr("Illigal opCode!", Vsm.cnt);
			goto Err;
			break;

		} // switch(op)
	} // for


	return;

End:
	Vsm.mode = MODE_END;
	return;

Err:
	Vsm.mode = MODE_ERR;
	return;
}
//---------------------------------------------------------------------------
bool VsmIsExecEnd()
{
	return (Vsm.mode == MODE_EXEC) ? FALSE: TRUE;
}
//---------------------------------------------------------------------------
void VsmCodeAdd(u16 code)
{
	if(VsmCodeIsFull() == TRUE)
	{
		return;
	}

	Vsm.codeBuf[Vsm.codeCnt++] = code;
}
//---------------------------------------------------------------------------
void VsmCodeAddEnd()
{
	TRACEOUT("End\n");

	VsmCodeAdd(CODE_END);
}
//---------------------------------------------------------------------------
void VsmCodeAddNext()
{
	TRACEOUT("Next\n");

	VsmCodeAdd(CODE_NEXT);
}
//---------------------------------------------------------------------------
void VsmCodeAddReturn()
{
	TRACEOUT("Return\n");

	VsmCodeAdd(CODE_RETURN);
}
//---------------------------------------------------------------------------
void VsmCodeAddCls()
{
	TRACEOUT("Cls\n");

	VsmCodeAdd(CODE_CLS);
}
//---------------------------------------------------------------------------
void VsmCodeAddGosub()
{
	TRACEOUT("Gosub\n");

	VsmCodeAdd(CODE_GOSUB);
}
//---------------------------------------------------------------------------
void VsmCodeAddGoto()
{
	TRACEOUT("Goto\n");

	VsmCodeAdd(CODE_GOTO);
}
//---------------------------------------------------------------------------
void VsmCodeAddRnd()
{
	TRACEOUT("Rnd\n");

	VsmCodeAdd(CODE_RND);
}
//---------------------------------------------------------------------------
void VsmCodeAddBeq()
{
	TRACEOUT("Beq(%d)\n", Vsm.codeCnt+1);

	VsmCodeAdd(CODE_BEQ);
	VsmCodeAdd(0);
}
//---------------------------------------------------------------------------
void VsmCodeAddJmp()
{
	TRACEOUT("Jmp(%d)\n", Vsm.codeCnt+1);

	VsmCodeAdd(CODE_JMP);
	VsmCodeAdd(0);
}
//---------------------------------------------------------------------------
void VsmCodeAddPset()
{
	TRACEOUT("Pset\n");

	VsmCodeAdd(CODE_PSET);
}
//---------------------------------------------------------------------------
void VsmCodeAddLine()
{
	TRACEOUT("Line\n");

	VsmCodeAdd(CODE_LINE);
}
//---------------------------------------------------------------------------
void VsmCodeAddCircle()
{
	TRACEOUT("Circle\n");

	VsmCodeAdd(CODE_CIRCLE);
}
//---------------------------------------------------------------------------
void VsmCodeAddForInit()
{
	TRACEOUT("ForInit\n");

	VsmCodeAdd(CODE_FOR_INIT);
}
//---------------------------------------------------------------------------
void VsmCodeAddFor()
{
	TRACEOUT("For\n");

	VsmCodeAdd(CODE_FOR);
	VsmCodeAdd(0);
}
//---------------------------------------------------------------------------
void VsmCodeAddPrint()
{
	TRACEOUT("Print\n");

	VsmCodeAdd(CODE_PRINT);
}
//---------------------------------------------------------------------------
void VsmCodeAddPrintLf()
{
	TRACEOUT("PrintLf\n");

	VsmCodeAdd(CODE_PRINT_LF);
}
//---------------------------------------------------------------------------
void VsmCodeAddVar()
{
	TRACEOUT("Var\n");

	VsmCodeAdd(CODE_VAR);
}
//---------------------------------------------------------------------------
void VsmCodeAddArray()
{
	TRACEOUT("Array\n");

	VsmCodeAdd(CODE_ARRAY);
}
//---------------------------------------------------------------------------
void VsmCodeAddDim()
{
	TRACEOUT("Dim\n");

	VsmCodeAdd(CODE_DIM);
}
//---------------------------------------------------------------------------
void VsmCodeAddAdr(u16 adr)
{
	TRACEOUT("------------------------------\n"); 
	TRACEOUT("Adr %d\n", adr);

	VsmAdrAdd(adr, VsmCodeGetCnt());

	VsmCodeAdd(CODE_ADR);
	VsmCodeAdd(adr);
}
//---------------------------------------------------------------------------
void VsmCodeAddGt()
{
	TRACEOUT("Gt\n");

	VsmCodeAdd(CODE_GT);
}
//---------------------------------------------------------------------------
void VsmCodeAddGe()
{
	TRACEOUT("Ge\n");

	VsmCodeAdd(CODE_GE);
}
//---------------------------------------------------------------------------
void VsmCodeAddLt()
{
	TRACEOUT("Lt\n");

	VsmCodeAdd(CODE_LT);
}
//---------------------------------------------------------------------------
void VsmCodeAddLe()
{
	TRACEOUT("Le\n");

	VsmCodeAdd(CODE_LE);
}
//---------------------------------------------------------------------------
void VsmCodeAddNe()
{
	TRACEOUT("Ne\n");

	VsmCodeAdd(CODE_NE);
}
//---------------------------------------------------------------------------
void VsmCodeAddEq()
{
	TRACEOUT("Eq\n");

	VsmCodeAdd(CODE_EQ);
}
//---------------------------------------------------------------------------
void VsmCodeAddPlus()
{
	TRACEOUT("Plus\n");

	VsmCodeAdd(CODE_PLUS);
}
//---------------------------------------------------------------------------
void VsmCodeAddMinus()
{
	TRACEOUT("Minus\n");

	VsmCodeAdd(CODE_MINUS);
}
//---------------------------------------------------------------------------
void VsmCodeAddMult()
{
	TRACEOUT("Mult\n");

	VsmCodeAdd(CODE_MULT);
}
//---------------------------------------------------------------------------
void VsmCodeAddDiv()
{
	TRACEOUT("Div\n");

	VsmCodeAdd(CODE_DIV);
}
//---------------------------------------------------------------------------
void VsmCodeAddMod()
{
	TRACEOUT("Mod\n");

	VsmCodeAdd(CODE_MOD);
}
//---------------------------------------------------------------------------
void VsmCodeAddPushNum(u16 num)
{
	TRACEOUT("PushNum %d\n", num);

	VsmCodeAdd(CODE_PUSH_NUM);
	VsmCodeAdd(num);
}
//---------------------------------------------------------------------------
void VsmCodeAddPushVar(char* pStr)
{
	u16 adr = VsmWorkAddStr(pStr);

	TRACEOUT("PushVar %d(%s)\n", adr, pStr);

	VsmCodeAdd(CODE_PUSH_VAR);
	VsmCodeAdd(adr);
}
//---------------------------------------------------------------------------
void VsmCodeAddPushArray(char* pStr)
{
	u16 adr = VsmWorkAddStr(pStr);

	TRACEOUT("PushArray %d(%s)\n", adr, pStr);

	VsmCodeAdd(CODE_PUSH_ARRAY);
	VsmCodeAdd(adr);
}
//---------------------------------------------------------------------------
void VsmCodeAddPushStr(char* pStr)
{
	u16 adr = VsmWorkAddStr(pStr);

	TRACEOUT("PushStr %d(%s)\n", adr, pStr);

	VsmCodeAdd(CODE_PUSH_STR);
	VsmCodeAdd(adr);
}
//---------------------------------------------------------------------------
void VsmCodePatchBeq(u16 tmpCnt)
{
	TRACEOUT("PatchBeq(%d, %d)\n", tmpCnt, Vsm.codeCnt);

	Vsm.codeBuf[tmpCnt] = Vsm.codeCnt;
}
//---------------------------------------------------------------------------
void VsmCodePatchJmp(u16 tmpCnt)
{
	TRACEOUT("PatchJmp(%d, %d)\n", tmpCnt, Vsm.codeCnt);

	Vsm.codeBuf[tmpCnt] = Vsm.codeCnt;
}
//---------------------------------------------------------------------------
void VsmCodePatchFor(u16 tmpCnt)
{
	TRACEOUT("PatchFor(%d, %d)\n", tmpCnt, Vsm.codeCnt);

	Vsm.codeBuf[tmpCnt] = Vsm.codeCnt;
}
//---------------------------------------------------------------------------
u16 VsmCodeGetCnt()
{
	return Vsm.codeCnt;
}
//---------------------------------------------------------------------------
bool VsmCodeIsFull()
{
	return (Vsm.codeCnt >= VSM_MAX_CODE_BUF_CNT) ? TRUE : FALSE;
}





//---------------------------------------------------------------------------
u16 VsmVarReg(u16 strAdr, u16 size)
{
	if(VsmVarIsFull() == TRUE)
	{
		return VSM_MAX_WORK_BUF_CNT-1;
	}

	ST_VSM_VAR_BUF* p = &Vsm.varBuf[Vsm.varCnt];

	p->type   = (size == 1) ? VAR_VAR : VAR_ARRAY;
	p->size   = size;
	p->strAdr = strAdr;
	p->varAdr = VsmWorkAddVar(size);

	Vsm.varCnt++;
	return p->varAdr;
}
//---------------------------------------------------------------------------
u16 VsmVarAdd(u16 strAdr)
{
	// ϐ̂݌Ăяoł܂iz͎Oɓo^ĂȂƃG[ɂȂ邽߁j

	if(VsmVarIsReg(strAdr, VAR_VAR) == TRUE)
	{
		return VsmVarGetAdr(strAdr, VAR_VAR);
	}

	return VsmVarReg(strAdr, 1);
}
//---------------------------------------------------------------------------
u16 VsmVarGetAdr(u16 strAdr, u16 type)
{
	u16 i;

	for(i=0; i<Vsm.varCnt; i++)
	{
		if(strAdr == Vsm.varBuf[i].strAdr && type == Vsm.varBuf[i].type)
		{
			return Vsm.varBuf[i].varAdr;
		}
	}

	_ASSERT(0 && "VsmVarGetAdr Err!");
	return VSM_MAX_WORK_BUF_CNT-1;
}
//---------------------------------------------------------------------------
u16 VsmVarGetSize(u16 strAdr, u16 type)
{
	u16 i;

	for(i=0; i<Vsm.varCnt; i++)
	{
		if(strAdr == Vsm.varBuf[i].strAdr && type == Vsm.varBuf[i].type)
		{
			return Vsm.varBuf[i].size;
		}
	}

	_ASSERT(0 && "VsmVarGetSize Err!");
	return VSM_MAX_WORK_BUF_CNT-1;
}
//---------------------------------------------------------------------------
bool VsmVarIsReg(u16 strAdr, u16 type)
{
	u16 i;

	for(i=0; i<Vsm.varCnt; i++)
	{
		if(strAdr == Vsm.varBuf[i].strAdr && type == Vsm.varBuf[i].type)
		{
			return TRUE;
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
bool VsmVarIsFull()
{
	return (Vsm.varCnt >= VSM_MAX_VAR_BUF_CNT) ? TRUE : FALSE;
}





//---------------------------------------------------------------------------
u16 VsmWorkAddStr(char* pStr)
{
	u16 len  = _Strlen(pStr);
	u16 size = (len + 1) / 2 + 1;	// workBufPʁiu16jɕϊ܂B㔼+1́A敪ł

	if(VsmWorkIsReg(pStr) == TRUE)
	{
		return VsmWrokGetAdr(pStr);
	}
	else if((Vsm.workCnt + size) >= VSM_MAX_WORK_BUF_CNT-1)
	{
		Vsm.workCnt += size;
		return VSM_MAX_WORK_BUF_CNT-1;
	}


	u16 ret = Vsm.workCnt;

	// obt@̃TCYi[܂
	Vsm.workBuf[Vsm.workCnt] = size;
	Vsm.workCnt += 1;

	// i[܂
	u8* pSrc = (u8*)pStr;
	u8* pDst = (u8*)&Vsm.workBuf[Vsm.workCnt];
	u16 i;

	for(i=0; i<len; i++)
	{
		*pDst++ = *pSrc++;
	}
	Vsm.workCnt += size;

	return ret;
}
//---------------------------------------------------------------------------
u16 VsmWorkAddVar(u16 size)
{
	if((Vsm.workCnt + size) >= VSM_MAX_WORK_BUF_CNT-1)
	{
		Vsm.workCnt += size;
		return VSM_MAX_WORK_BUF_CNT-1;
	}

	u16 ret = Vsm.workCnt;
	Vsm.workCnt += size;
	return ret;
}
//---------------------------------------------------------------------------
u16 VsmWrokGetAdr(char* pStr)
{
	s16 cnt = 0;

	while(cnt < Vsm.workCnt)
	{
		if(_Strncmp(pStr, (char*)&Vsm.workBuf[cnt+1], Vsm.workBuf[cnt]*2) == 0)
		{
			return cnt;
		}

		cnt += Vsm.workBuf[cnt] + 1;
	}

	_ASSERT(0 && "VsmWrokGetAdr Err!");
	return VSM_MAX_WORK_BUF_CNT-1;
}
//---------------------------------------------------------------------------
bool VsmWorkIsFull()
{
	return (Vsm.workCnt >= VSM_MAX_WORK_BUF_CNT) ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
bool VsmWorkIsReg(char* pStr)
{
	s16 cnt = 0;

	while(cnt < Vsm.workCnt)
	{
		if(_Strncmp(pStr, (char*)&Vsm.workBuf[cnt+1], Vsm.workBuf[cnt]*2) == 0)
		{
			return TRUE;
		}

		cnt += Vsm.workBuf[cnt] + 1;
	}

	return FALSE;
}





//---------------------------------------------------------------------------
void VsmStackPush(u16 type, u16 num)
{
	if(VsmStackIsFull() == TRUE)
	{
		return;
	}

	TRACEOUT("    Push[%4d] - %d\n", Vsm.stackCnt+0, num);
	TRACEOUT("    Push[%4d] - %d\n", Vsm.stackCnt+1, type);
	Vsm.stackBuf[ Vsm.stackCnt++ ] = num;
	Vsm.stackBuf[ Vsm.stackCnt++ ] = type;
}
//---------------------------------------------------------------------------
u16 VsmStackPop()
{
	_ASSERT(Vsm.stackCnt > 0);

	TRACEOUT("    Pop [%4d] - %d\n", Vsm.stackCnt-1, Vsm.stackBuf[Vsm.stackCnt-1]);
	return Vsm.stackBuf[ --Vsm.stackCnt ];
}
//---------------------------------------------------------------------------
u16 VsmStackPopNum()
{
	u16 type = VsmStackPop();
	u16 num  = VsmStackPop();

	if(type == STACK_NUM)
	{
		return num;
	}
	else if(type == STACK_VAR)
	{
		return Vsm.workBuf[ VsmVarAdd(num) ];
	}
	else if(type == STACK_ARRAY)
	{
		// TODO ċAĂ܂

		// o^`FbN
		if(VsmVarIsReg(num, VAR_ARRAY) == FALSE)
		{
			return Vsm.workBuf[VSM_MAX_WORK_BUF_CNT-1];
		}

		// ͈̓`FbN
		u16 num2 = VsmStackPopNum();
		if(VsmVarGetSize(num, VAR_ARRAY) <= num2)
		{
			return Vsm.workBuf[VSM_MAX_WORK_BUF_CNT-1];
		}

		return Vsm.workBuf[VsmVarGetAdr(num, VAR_ARRAY) + num2];
	}

	_ASSERT(0 && "VsmStackPopNum Err!");
	return 0;
}
//---------------------------------------------------------------------------
char* VsmStackPopStr()
{
	u16 type = VsmStackPop();
	u16 num  = VsmStackPop();

	if(type == STACK_STR)
	{
		return (char*)&Vsm.workBuf[num + 1];
	}

	_ASSERT(0 && "VsmStackPopStr Err!");
	return (char*)&Vsm.workBuf[0];
}
//---------------------------------------------------------------------------
u16 VsmStackPopAdr()
{
	_ASSERT(Vsm.stackCnt >= 2);

	u16 type = VsmStackPop();
	u16 num  = VsmStackPop();

	if(type == STACK_VAR || type == STACK_ARRAY)
	{
		return VsmVarAdd(num);
	}

	_ASSERT(0 && "VsmStackPopAdr Err!");
	return 0;
}
//---------------------------------------------------------------------------
u16 VsmStackGetTopType()
{
	if(Vsm.stackCnt <= 0)
	{
		return STACK_NONE;
	}

	return Vsm.stackBuf[ Vsm.stackCnt-1 ];
}
//---------------------------------------------------------------------------
u16 VsmStackGetTopNum()
{
	if(Vsm.stackCnt <= 1)
	{
		return STACK_NONE;
	}

	return Vsm.stackBuf[ Vsm.stackCnt-2 ];
}
//---------------------------------------------------------------------------
bool VsmStackIsFull()
{
	return (Vsm.stackCnt+2 >= VSM_MAX_STACK_BUF_CNT) ? TRUE : FALSE;
}





//---------------------------------------------------------------------------
void VsmAdrAdd(u16 adr, u16 pc)
{
	if(VsmAdrIsFull() == TRUE)
	{
		return;
	}

	Vsm.adrBuf[Vsm.adrCnt].adr = adr;
	Vsm.adrBuf[Vsm.adrCnt].pc  = pc;
	Vsm.adrCnt++;
}
//---------------------------------------------------------------------------
u16 VsmAdrGetPc(u16 adr)
{
	u16 i;

	for(i=0; i<Vsm.adrCnt; i++)
	{
		if(adr == Vsm.adrBuf[i].adr)
		{
			return Vsm.adrBuf[i].pc;
		}
	}

	_ASSERT(0 && "VsmAdrGetPc Err!");
	return 0;
}
//---------------------------------------------------------------------------
bool VsmAdrIsPc(u16 adr)
{
	u16 i;

	for(i=0; i<Vsm.adrCnt; i++)
	{
		if(adr == Vsm.adrBuf[i].adr)
		{
			return TRUE;
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
bool VsmAdrIsFull()
{
	return (Vsm.adrCnt >= VSM_MAX_ADR_BUF_CNT) ? TRUE : FALSE;
}





//---------------------------------------------------------------------------
void VsmDebugCodeBuf()
{
	TRACEOUT("------------------------------\n"); 
	TRACEOUT("DebugCodeBuf\n");

	u16 i;

	for(i=0; i<Vsm.codeCnt; i++)
	{
		TRACEOUT("[%4d] - %4d\n", i, Vsm.codeBuf[i]);
	}

	TRACEOUT("\nTOTAL: %d\n", Vsm.codeCnt);
}
//---------------------------------------------------------------------------
void VsmDebugWorkBuf()
{
	TRACEOUT("------------------------------\n"); 
	TRACEOUT("DebugWorkBuf\n");

	u16 i;

	for(i=0; i<Vsm.workCnt; i++)
	{
		TRACEOUT("[%4d] - %4x\n", i, Vsm.workBuf[i]);
	}

	TRACEOUT("\nTOTAL: %d\n", Vsm.workCnt);
}
//---------------------------------------------------------------------------
void VsmDebugAdrBuf()
{
	TRACEOUT("------------------------------\n"); 
	TRACEOUT("DebugWorkBuf\n");

	u16 i;

	for(i=0; i<Vsm.adrCnt; i++)
	{
		TRACEOUT("%4d - %4d\n", Vsm.adrBuf[i].adr , Vsm.adrBuf[i].pc);
	}

	TRACEOUT("\nTOTAL: %d\n", Vsm.adrCnt);
}
